from decimal import Decimal
from fractions import Fraction
from itertools import count, islice
from Z0Z_tools import updateExtendPolishDictionaryLists, stringItUp
import datetime
import numpy
import unittest
import uuid

class TestStringItUp(unittest.TestCase):

    def test_empty_input(self):
        self.assertEqual(stringItUp(), [])

    def test_with_bytearray(self):
        self.assertEqual(stringItUp(bytearray(b"bytearray")), ["bytearray(b'bytearray')"])

    def test_with_generator(self):
        def gen():
            for i in range(3):
                yield i
        self.assertEqual(stringItUp(gen()), ['0', '1', '2'])

    def test_with_infinite_iterator(self):
        infinite_gen = count()
        limited_gen = islice(infinite_gen, 5)
        self.assertEqual(stringItUp(limited_gen), ['0', '1', '2', '3', '4'])

    def test_with_object_with_custom_iter(self):
        class CustomIter:
            def __iter__(self):
                return iter([1, 2, 3])
        self.assertEqual(stringItUp(CustomIter()), ['1', '2', '3'])

    def test_with_recursive_structure(self):
        chicken_tikka_masala = []
        chicken_tikka_masala.append(chicken_tikka_masala)
        self.assertEqual(stringItUp(chicken_tikka_masala), ['[[...]]'])

    def test_with_recursive_structure2(self):
        chicken_tikka_masala = []
        chicken_tikka_masala.append(chicken_tikka_masala)
        self.assertEqual(stringItUp([chicken_tikka_masala, 'wazzup']), ["[[[...]], 'wazzup']"])

    def test_with_nan_and_inf(self):
        self.assertEqual(stringItUp(float('nan'), float('inf')), ['nan', 'inf'])

    def test_with_large_numbers(self):
        large_number = 10**100
        self.assertEqual(stringItUp(large_number), [str(large_number)])

    def test_with_decimal(self):
        self.assertEqual(stringItUp(Decimal('1.1')), ['1.1'])

    def test_with_fraction(self):
        self.assertEqual(stringItUp(Fraction(1, 3)), ['1/3'])

    def test_with_dates(self):
        date = datetime.date(2021, 1, 1)
        self.assertEqual(stringItUp(date), ['2021-01-01'])

    def test_with_times(self):
        time = datetime.time(12, 34, 56)
        self.assertEqual(stringItUp(time), ['12:34:56'])

    def test_with_datetime(self):
        dt = datetime.datetime(2021, 1, 1, 12, 34, 56)
        self.assertEqual(stringItUp(dt), ['2021-01-01 12:34:56'])

    def test_with_uuid(self):
        my_uuid = uuid.uuid4()
        self.assertEqual(stringItUp(my_uuid), [str(my_uuid)])

    def test_with_memoryview(self):
        result = stringItUp(memoryview(b"memoryview"))
        expected_prefix = "<memory at 0x"
        self.assertTrue(result[0].startswith(expected_prefix))

    def test_empty_iterable_types(self):
        self.assertEqual(stringItUp([], (), set()), [])

    def test_mixed_nested_iterables(self):
        data = [1, (2, {3, "four"}), {"five": [6, 7]}]
        self.assertCountEqual(stringItUp(data), ["1", "2", "3", "four", "five", "6", "7"])

    def test_large_data(self):
        large_list = list(range(1000))
        result = stringItUp(large_list)
        self.assertEqual(len(result), 1000)
        for i in range(1000):
            self.assertEqual(result[i], str(i))

class TestUpdateExtendDictionaryLists(unittest.TestCase):

    def test_with_mixed_types_in_values(self):
        primus = {'a': [1, 'two'], 'b': [True, None]}
        secundus = {'a': [3.14, 'four'], 'b': [False, 'none']}
        expected = {
            'a': [1, 'two', 3.14, 'four'],
            'b': [True, None, False, 'none']
        }
        result = updateExtendPolishDictionaryLists(primus, secundus)
        self.assertEqual(result, expected)

    def test_with_non_string_keys(self):
        primus = {None: [3], True: [2]}
        secundus = {1: [1], (4, 5): [4]}
        expected ={'(4, 5)': [4], '1': [1], 'None': [3], 'True': [2]}
        result = updateExtendPolishDictionaryLists(primus, secundus) # type: ignore
        self.assertEqual(result, expected)

    def test_with_conflicting_data_types(self):
        primus = {'a': 1, 'b': 2}
        secundus = {'a': 3, 'c': 4}
        with self.assertRaises(TypeError):
            result = updateExtendPolishDictionaryLists(primus, secundus) # type: ignore

    def test_with_kill_erroneous_data_types(self):
        primus = {'a': [1, 2], 'b': [3, 4]}
        secundus = {'a': 3, 'c': 4}
        expected = {'a': [1, 2], 'b': [3, 4]}
        result = updateExtendPolishDictionaryLists(primus, secundus, killErroneousDataTypes=True) # type: ignore
        self.assertEqual(result, expected)

    def test_basic_functionality(self):
        primus = {'a': [3, 1], 'b': [2]}
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [3, 1, 9, 6, 1, 22, 3], 'b': [2, 111111, 2, 3]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_ignore_list_ordering(self):
        primus = {'a': [3, 1], 'b': [2]}
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [1, 1, 3, 3, 6, 9, 22], 'b': [2, 2, 3, 111111]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=True)
        self.assertEqual(result, expected)

    def test_destroy_duplicates(self):
        primus = {'a': [3, 1], 'b': [2]}
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [3, 1, 9, 6, 22], 'b': [2, 111111, 3]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=True, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus2(self):
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        result = updateExtendPolishDictionaryLists({}, secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus_ignore_list_ordering2(self):
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [1, 3, 6, 9, 22], 'b': [2, 3, 111111]}
        result = updateExtendPolishDictionaryLists({}, secundus, destroyDuplicates=False, reorderLists=True)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus_destroy_duplicates2(self):
        secundus = {'a': [9, 6, 1, 22, 3, 9], 'b': [111111, 2, 3, 2]}
        expected = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        result = updateExtendPolishDictionaryLists({}, secundus, destroyDuplicates=True, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus_destroy_duplicates_ignore_list_ordering2(self):
        secundus = {'a': [9, 6, 1, 22, 3, 9], 'b': [111111, 2, 3, 2]}
        expected = {'a': [1, 3, 6, 9, 22], 'b': [2, 3, 111111]}
        result = updateExtendPolishDictionaryLists({}, secundus, destroyDuplicates=True, reorderLists=True)
        self.assertEqual(result, expected)

    def test_empty_dictionaries(self):
        primus = {}
        secundus = {}
        expected = {}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_empty_and_non_empty_dictionary(self):
        primus = {}
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_non_empty_and_empty_dictionary(self):
        primus = {'a': [3, 1], 'b': [2]}
        secundus = {}
        expected = {'a': [3, 1], 'b': [2]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_empty_dictionaries_with_options(self):
        primus = {}
        secundus = {}
        expected = {}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=True, reorderLists=True)
        self.assertEqual(result, expected)

    def test_single_empty_dictionary(self):
        primus = {}
        expected = {}
        result = updateExtendPolishDictionaryLists(primus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_empty_dictionary_with_options(self):
        primus = {}
        expected = {}
        result = updateExtendPolishDictionaryLists(primus, destroyDuplicates=True, reorderLists=True)
        self.assertEqual(result, expected)

    def test_single_dictionary_primus(self):
        primus = {'a': [3, 1], 'b': [2]}
        expected = {'a': [3, 1], 'b': [2]}
        result = updateExtendPolishDictionaryLists(primus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_primus_ignore_list_ordering_variant(self):
        primus = {'a': [3, 1], 'b': [2]}
        expected = {'a': [1, 3], 'b': [2]}
        result = updateExtendPolishDictionaryLists(primus, destroyDuplicates=False, reorderLists=True)
        self.assertEqual(result, expected)

    def test_single_dictionary_primus_destroy_duplicates_ignore_list_ordering_variant(self):
        primus = {'a': [3, 1, 3], 'b': [2, 2]}
        expected = {'a': [3, 1], 'b': [2]}
        result = updateExtendPolishDictionaryLists(primus, destroyDuplicates=True, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_primus_destroy_duplicates_ignore_list_ordering(self):
        primus = {'a': [3, 1, 3], 'b': [2, 2]}
        expected = {'a': [1, 3], 'b': [2]}
        result = updateExtendPolishDictionaryLists(primus, destroyDuplicates=True, reorderLists=True)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus(self):
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        result = updateExtendPolishDictionaryLists(secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus_ignore_list_ordering(self):
        secundus = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        expected = {'a': [1, 3, 6, 9, 22], 'b': [2, 3, 111111]}
        result = updateExtendPolishDictionaryLists(secundus, destroyDuplicates=False, reorderLists=True)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus_destroy_duplicates(self):
        secundus = {'a': [9, 6, 1, 22, 3, 9], 'b': [111111, 2, 3, 2]}
        expected = {'a': [9, 6, 1, 22, 3], 'b': [111111, 2, 3]}
        result = updateExtendPolishDictionaryLists(secundus, destroyDuplicates=True, reorderLists=False)
        self.assertEqual(result, expected)

    def test_single_dictionary_secundus_destroy_duplicates_ignore_list_ordering(self):
        secundus = {'a': [9, 6, 1, 22, 3, 9], 'b': [111111, 2, 3, 2]}
        expected = {'a': [1, 3, 6, 9, 22], 'b': [2, 3, 111111]}
        result = updateExtendPolishDictionaryLists(secundus, destroyDuplicates=True, reorderLists=True)
        self.assertEqual(result, expected)

    def test_with_sets(self):
        primus = {'a': {3, 1}, 'b': {2}}
        secundus = {'a': {9, 6, 1, 22, 3}, 'b': {111111, 2, 3}}
        expected = {'a': [3, 1, 9, 6, 22], 'b': [2, 111111, 3]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=True, reorderLists=False) # type: ignore
        self.assertCountEqual(result['a'], expected['a'])

    def test_with_tuples(self):
        primus = {'a': (3, 1), 'b': (2,)}
        secundus = {'a': (9, 6, 1, 22, 3), 'b': (111111, 2, 3)}
        expected = {'a': [3, 1, 9, 6, 1, 22, 3], 'b': [2, 111111, 2, 3]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=False)
        self.assertEqual(result, expected)

    def test_with_ndarray(self):
        primus = {'a': numpy.array([3, 1]), 'b': numpy.array([2])}
        secundus = {'a': numpy.array([9, 6, 1, 22, 3]), 'b': numpy.array([111111, 2, 3])}
        expected = {'a': [3, 1, 9, 6, 1, 22, 3], 'b': [2, 111111, 2, 3]}
        result = updateExtendPolishDictionaryLists(primus, secundus, destroyDuplicates=False, reorderLists=False) # type: ignore
        self.assertEqual(result, expected)

    def test_empty_input(self):
        self.assertEqual(stringItUp(), [])

    def test_single_string(self):
        self.assertEqual(stringItUp("hello"), ["hello"])

    def test_single_integer(self):
        self.assertEqual(stringItUp(123), ["123"])

    def test_single_float(self):
        self.assertEqual(stringItUp(3.14), ["3.14"])

    def test_single_list(self):
        self.assertCountEqual(stringItUp([1, 2, "three"]), ["1", "2", "three"])

    def test_single_tuple(self):
        self.assertCountEqual(stringItUp((1, 2, "three")), ["1", "2", "three"])

    def test_single_set(self):
        self.assertCountEqual(stringItUp({1, 2, "three"}), ["1", "2", "three"])

    def test_single_dict(self):
        self.assertCountEqual(stringItUp({"a": 1, "b": "two"}), ["a", "1", "b", "two"])

    def test_nested_list(self):
        self.assertCountEqual(stringItUp([1, [2, "three"], 4]), ["1", "2", "three", "4"])

    def test_nested_tuple(self):
        self.assertCountEqual(stringItUp((1, (2, "three"), 4)), ["1", "2", "three", "4"])

    def test_nested_set(self):
        self.assertCountEqual(stringItUp({1, frozenset({2, "three"}), 4}), ["1", "2", "three", "4"])

    def test_nested_dict(self):
        self.assertCountEqual(stringItUp({"a": 1, "b": {"c": 2, "d": "three"}}), ["a", "1", "b", "c", "2", "d", "three"])

    def test_mixed_data_types(self):
        self.assertCountEqual(stringItUp(1, "two", [3, "four"], {"five": 5}), ["1", "two", "3", "four", "five", "5"])

    def test_with_numpy_array(self):
        self.assertCountEqual(stringItUp(numpy.array([1, 2, 3])), ["1", "2", "3"])

    def test_with_custom_object(self):
        class MyObject:
            def __str__(self):
                return "MyObject"
        self.assertEqual(stringItUp(MyObject()), ["MyObject"])

    def test_with_none(self):
        self.assertEqual(stringItUp(None), ["None"])

    def test_with_boolean(self):
        self.assertCountEqual(stringItUp(True, False), ["True", "False"])

    def test_with_complex_number(self):
        self.assertEqual(stringItUp(1+2j), ["(1+2j)"])

    def test_with_bytes(self):
        self.assertEqual(stringItUp(b"bytes"), ["b'bytes'"])

if __name__ == '__main__':
    unittest.main()


from Z0Z_tools import readAudioFile, loadWaveforms, writeWav
from Z0Z_tools.ioAudio import resampleWaveform
from numpy.typing import NDArray
import numpy
import pathlib
import soundfile
import unittest
import tempfile
import os

# Define test data directory and file paths
pathDirectoryTestData = pathlib.Path("unittests/dataSamples")

pathFilenameAudioMono = pathDirectoryTestData / "testWooWooMono16kHz32integerClipping9sec.wav"
pathFilenameAudioStereo = pathDirectoryTestData / "testSine2ch5sec.wav"
pathFilenameNonAudioVideo = pathDirectoryTestData / "testVideo11sec.mkv"

listPathFilenamesAudioMonoCopies = [
    pathDirectoryTestData / f"testWooWooMono16kHz32integerClipping9secCopy{i}.wav" for i in range(1, 4)
]
listPathFilenamesAudioStereoCopies = [
    pathDirectoryTestData / f"testSine2ch5secCopy{i}.wav" for i in range(1, 5)
]

class TestReadAudioFile(unittest.TestCase):

    def setUp(self):
        self.pathFilenameAudioMono = pathFilenameAudioMono
        self.pathFilenameAudioStereo = pathFilenameAudioStereo
        self.pathFilenameNonAudioVideo = pathFilenameNonAudioVideo

    def test_read_mono_audio_file(self):
        waveform = readAudioFile(self.pathFilenameAudioMono)
        self.assertIsInstance(waveform, numpy.ndarray)
        self.assertEqual(waveform.ndim, 2)  # Mono should be converted to stereo, hence 2 dimensions
        self.assertEqual(waveform.shape[0], 2) # Verify stereo output

    def test_read_stereo_audio_file(self):
        waveform = readAudioFile(self.pathFilenameAudioStereo)
        self.assertIsInstance(waveform, numpy.ndarray)
        self.assertEqual(waveform.ndim, 2)
        self.assertEqual(waveform.shape[0], 2)

    def test_read_audio_file_resampling(self):
        waveform_original_sr = readAudioFile(self.pathFilenameAudioMono, sampleRate=16000) # Original Sample Rate
        waveform_resampled_sr = readAudioFile(self.pathFilenameAudioMono, sampleRate=44100) # Resampled
        self.assertNotEqual(waveform_original_sr.shape[1], waveform_resampled_sr.shape[1])


    def test_invalid_file(self):
        with self.assertRaises(FileNotFoundError): # Or appropriate exception from soundfile library
            readAudioFile("nonexistent_file.wav")
        with self.assertRaises(soundfile.LibsndfileError): # Or a more specific exception type
            readAudioFile(self.pathFilenameNonAudioVideo)


class TestLoadWaveforms(unittest.TestCase):

    def setUp(self):
        self.listPathFilenamesAudioMonoCopies = listPathFilenamesAudioMonoCopies
        self.listPathFilenamesAudioStereoCopies = listPathFilenamesAudioStereoCopies

    def test_load_waveforms_mono(self):
        array_waveforms = loadWaveforms(self.listPathFilenamesAudioMonoCopies, sampleRate=44100)
        self.assertEqual(array_waveforms.shape, (2, 396900, 3))

    def test_load_waveforms_stereo(self):
        array_waveforms = loadWaveforms(self.listPathFilenamesAudioStereoCopies, sampleRate=44100)
        self.assertEqual(array_waveforms.shape, (2, 220500, 4))

    def test_load_waveforms_mixed_channels(self):
        mixed_files = self.listPathFilenamesAudioMonoCopies[:1] + self.listPathFilenamesAudioStereoCopies[:1]
        array_waveforms = loadWaveforms(mixed_files, sampleRate=44100)
        self.assertEqual(array_waveforms.shape, (2, 396900, 2))

    def test_load_waveforms_empty_list(self):
        with self.assertRaises(ValueError): # Or a more appropriate exception
            loadWaveforms([])

    def test_load_waveforms_invalid_file(self):
        invalid_files = self.listPathFilenamesAudioMonoCopies + ["invalid_file.wav"]
        with self.assertRaises(FileNotFoundError) as context: # Or a more specific exception
            loadWaveforms(invalid_files) # type: ignore
        # Optionally check the error message for more precise testing

class TestResampleWaveform(unittest.TestCase):

    def setUp(self):
        self.pathFilenameAudioMono = pathFilenameAudioMono
        self.pathFilenameAudioStereo = pathFilenameAudioStereo

        # Load real audio files
        self.arrayWaveformMono: NDArray[numpy.float32]
        self.sampleRateMono: int
        self.arrayWaveformMono, self.sampleRateMono = soundfile.read(self.pathFilenameAudioMono, dtype='float32') # type: ignore
        self.arrayWaveformMono = self.arrayWaveformMono.astype(numpy.float32)

        self.arrayWaveformStereo: NDArray[numpy.float32]
        self.sampleRateStereo: int
        self.arrayWaveformStereo, self.sampleRateStereo = soundfile.read(self.pathFilenameAudioStereo, dtype='float32') # type: ignore
        self.arrayWaveformStereo = self.arrayWaveformStereo.astype(numpy.float32)

    def testResampleWaveformUpsampleMono(self):
        """
        Test resampling a mono waveform from a lower sample rate to a higher sample rate.
        """
        sampleRateDesired: int = 44100
        arrayWaveformResampled: NDArray[numpy.float32] = resampleWaveform(
            waveform=self.arrayWaveformMono,
            sampleRateDesired=sampleRateDesired,
            sampleRateSource=self.sampleRateMono
        )
        countSamplesExpected: int = int(self.arrayWaveformMono.shape[0] * (sampleRateDesired / self.sampleRateMono))
        self.assertEqual(arrayWaveformResampled.shape[0], countSamplesExpected)

    def testResampleWaveformDownsampleStereo(self):
        """
        Test resampling a stereo waveform from a higher sample rate to a lower sample rate.
        """
        sampleRateDesired: int = 22050
        arrayWaveformResampled: NDArray[numpy.float32] = resampleWaveform(
            waveform=self.arrayWaveformStereo,
            sampleRateDesired=sampleRateDesired,
            sampleRateSource=self.sampleRateStereo
        )
        countSamplesExpected: int = int(self.arrayWaveformStereo.shape[0] * (sampleRateDesired / self.sampleRateStereo))
        self.assertEqual(arrayWaveformResampled.shape[0], countSamplesExpected)

    def testResampleWaveformEqualSampleRate(self):
        """
        Test that the waveform is unchanged if sample rates are equal.
        """
        arrayWaveformResampled: NDArray[numpy.float32] = resampleWaveform(
            waveform=self.arrayWaveformStereo,
            sampleRateDesired=self.sampleRateStereo,
            sampleRateSource=self.sampleRateStereo
        )
        self.assertTrue(numpy.array_equal(arrayWaveformResampled, self.arrayWaveformStereo))

    def testResampleWaveformInvalidInput(self):
        """
        Test resampling with invalid input types.
        """
        with self.assertRaises(AttributeError):
            resampleWaveform('invalidInput', 44100, 22050) # type: ignore

    def testResampleWaveformNegativeSampleRate(self):
        """
        Test resampling with a negative sample rate.
        """
        with self.assertRaises(ValueError):
            resampleWaveform(self.arrayWaveformStereo, -44100, self.sampleRateStereo)

class TestWriteWav(unittest.TestCase):
    def test_write_wav_mono(self):
        waveform = numpy.random.rand(1, 1000)  #Example mono waveform
        with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as temp_file:
            filepath = temp_file.name
            writeWav(filepath, waveform)
            self.assertTrue(os.path.exists(filepath))
            # Add assertions to check file content if necessary (using soundfile to read it back)

    def test_write_wav_stereo(self):
        waveform = numpy.random.rand(2, 1000) #Example stereo waveform
        with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as temp_file:
            filepath = temp_file.name
            writeWav(filepath, waveform)
            self.assertTrue(os.path.exists(filepath))
            # Add assertions to check file content

    def test_write_wav_directory_creation(self):
        waveform = numpy.random.rand(2,1000)
        temp_dir = tempfile.mkdtemp()
        filepath = os.path.join(temp_dir, "test.wav")
        writeWav(filepath, waveform)
        self.assertTrue(os.path.exists(filepath))


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path
import tempfile
import unittest
from Z0Z_tools.Z0Z_io import dataTabularTOpathFilenameDelimited, findRelativePath

class TestDataTabularTOpathFilenameDelimited(unittest.TestCase):
    """Test suite for dataTabularTOpathFilenameDelimited function."""

    def setUp(self):
        """Create a temporary directory and test files for each test."""
        self.directoryTemporary = tempfile.mkdtemp()
        self.pathDirectory = Path(self.directoryTemporary)
        self.pathFilenameTest = self.pathDirectory / 'test_output.txt'
        
        # Test data that multiple tests will use
        self.tableColumnsBasic = ['String', 'Integer', 'Float']
        self.tableRowsBasic = [
            ['apple', 1, 1.5],
            ['banana', 2, 2.5],
            ['cherry', 3, 3.5]
        ]

    def tearDown(self):
        """Clean up temporary files after each test."""
        import shutil
        try:
            if self.pathFilenameTest.exists():
                self.pathFilenameTest.unlink()
            shutil.rmtree(self.directoryTemporary)
        except Exception as ERRORmessage:
            print(f"Warning: Cleanup failed - {ERRORmessage}")

    def verifyFileContent(self, pathFilename: Path, expectedContent: str) -> None:
        """Helper method to verify file content."""
        self.assertTrue(pathFilename.exists(), f"File not created: {pathFilename}")
        self.assertTrue(pathFilename.is_file(), f"Not a file: {pathFilename}")
        
        try:
            with open(pathFilename, 'r') as readStream:
                content = readStream.read().strip()
            self.assertEqual(content, expectedContent.strip())
        except Exception as ERRORmessage:
            self.fail(f"Failed to verify file content: {ERRORmessage}")

    def test_basicFunctionality(self):
        """Test basic functionality with different data types."""
        dataTabularTOpathFilenameDelimited(
            self.pathFilenameTest, 
            self.tableRowsBasic, 
            self.tableColumnsBasic
        )
        
        expectedContent = (
            'String\tInteger\tFloat\n'
            'apple\t1\t1.5\n'
            'banana\t2\t2.5\n'
            'cherry\t3\t3.5'
        )
        
        self.verifyFileContent(self.pathFilenameTest, expectedContent)

    def test_differentDelimiters(self):
        """Test using different delimiters."""
        tableRows = [['A', 'B'], ['C', 'D']]
        tableColumns = ['Column1', 'Column2']

        for delimiterOutput in [',', ';', '^|']:
            pathFilenameTest = self.pathDirectory / f'test_delimiterOutput.csv'
            
            dataTabularTOpathFilenameDelimited(
                pathFilenameTest, 
                tableRows, 
                tableColumns, 
                delimiterOutput
            )
            
            expectedContent = (
                f'Column1{delimiterOutput}Column2\n'
                f'A{delimiterOutput}B\n'
                f'C{delimiterOutput}D'
            )
            
            self.verifyFileContent(pathFilenameTest, expectedContent)

    def test_emptyData(self):
        """Test handling of empty data."""
        # Test with empty rows
        dataTabularTOpathFilenameDelimited(
            self.pathFilenameTest, 
            [], 
            ['Column1', 'Column2']
        )
        self.verifyFileContent(self.pathFilenameTest, 'Column1\tColumn2')

        # Test with empty columns
        dataTabularTOpathFilenameDelimited(
            self.pathFilenameTest, 
            [['A', 'B']], 
            []
        )
        self.verifyFileContent(self.pathFilenameTest, 'A\tB')

    def test_filePermissions(self):
        """Test handling of file permission errors."""
        if os.name != 'nt':  # Skip on Windows
            readOnlyDir = self.pathDirectory / 'readonly'
            readOnlyDir.mkdir(mode=0o444)
            pathFilenameReadOnly = readOnlyDir / 'test.txt'
            
            with self.assertRaises(PermissionError):
                dataTabularTOpathFilenameDelimited(
                    pathFilenameReadOnly, 
                    self.tableRowsBasic, 
                    self.tableColumnsBasic
                )

class TestFindRelativePath(unittest.TestCase):
    """Test suite for findRelativePath function."""

    def test_sameDirectory(self):
        """Test paths in the same directory."""
        pathSource = Path('/a/b/c/file1.txt')
        pathDestination = Path('/a/b/c/file2.txt')
        self.assertEqual(findRelativePath(pathSource, pathDestination), 'file2.txt')

    def test_parentChild(self):
        """Test parent to child directory relationships."""
        pathSource = Path('/a/b/c')
        pathDestination = Path('/a/b/c/d/e')
        self.assertEqual(findRelativePath(pathSource, pathDestination), 'd/e')

        # Child to parent
        self.assertEqual(findRelativePath(pathDestination, pathSource), '../..')

    def test_siblings(self):
        """Test sibling directory relationships."""
        pathSource = Path('/a/b/c1/d')
        pathDestination = Path('/a/b/c2/e')
        self.assertEqual(findRelativePath(pathSource, pathDestination), '../../c2/e')

    def test_differentBranches(self):
        """Test paths on completely different branches."""
        pathSource = Path('/a/b/c/d')
        pathDestination = Path('/w/x/y/z')
        self.assertEqual(findRelativePath(pathSource, pathDestination), '../../../../w/x/y/z')

    def test_stringInputs(self):
        """Test with string inputs instead of Path objects."""
        pathSource = '/a/b/c'
        pathDestination = '/a/b/d'
        self.assertEqual(findRelativePath(pathSource, pathDestination), '../d')

    def test_mixedInputTypes(self):
        """Test with mixed input types (string and Path)."""
        pathSource = '/a/b/c'
        pathDestination = Path('/a/b/d')
        self.assertEqual(findRelativePath(pathSource, pathDestination), '../d')
        self.assertEqual(findRelativePath(pathDestination, pathSource), '../c')

if __name__ == '__main__':
    unittest.main()


from typing import Optional, Union
import unittest
from unittest.mock import patch
from Z0Z_tools import defineConcurrencyLimit, oopsieKwargsie

class TestOopsieKwargsie(unittest.TestCase):
    """Test suite for oopsieKwargsie function."""

    def test_trueVariants(self):
        """Test various string representations of True."""
        trueVariants = ['True', 'TRUE', ' true ', 'TrUe']
        for variant in trueVariants:
            self.assertTrue(oopsieKwargsie(variant))

    def test_falseVariants(self):
        """Test various string representations of False."""
        falseVariants = ['False', 'FALSE', ' false ', 'FaLsE']
        for variant in falseVariants:
            self.assertFalse(oopsieKwargsie(variant))

    def test_noneVariants(self):
        """Test various string representations of None."""
        noneVariants = ['None', 'NONE', ' none ', 'NoNe']
        for variant in noneVariants:
            self.assertIsNone(oopsieKwargsie(variant))

    def test_otherStrings(self):
        """Test strings that should be returned unchanged."""
        testStrings = ['hello', '123', 'True story', 'False alarm']
        for testString in testStrings:
            self.assertEqual(oopsieKwargsie(testString), testString)

class TestDefineConcurrencyLimit(unittest.TestCase):
    """Test suite for defineConcurrencyLimit function."""

    def setUp(self):
        """Set up test fixtures."""
        self.cpuCountMocked = 8  # Define fixed CPU count for testing

    def runTestWithCPUcount(self, limitTest: Optional[Union[int, float, bool]], expectedResult: int) -> None:
        """Helper method to run tests with mocked CPU count."""
        with patch('multiprocessing.cpu_count', return_value=self.cpuCountMocked):
            resultTest = defineConcurrencyLimit(limitTest)
            self.assertEqual(resultTest, expectedResult)

    def test_defaultValues(self):
        """Test behavior with None, False, and 0."""
        for limitTest in [None, False, 0]:
            self.runTestWithCPUcount(limitTest, self.cpuCountMocked)

    def test_directInteger(self):
        """Test with direct integer values >= 1."""
        testCases = [
            (1, 1),
            (4, 4),
            (16, 16)
        ]
        for limitTest, expectedResult in testCases:
            self.runTestWithCPUcount(limitTest, expectedResult)

    def test_fractionFloat(self):
        """Test with float values between 0 and 1."""
        testCases = [
            (0.5, 4),  # 8 * 0.5 = 4
            (0.25, 2),  # 8 * 0.25 = 2
            (0.75, 6)  # 8 * 0.75 = 6
        ]
        for limitTest, expectedResult in testCases:
            self.runTestWithCPUcount(limitTest, expectedResult)

    def test_negativeFraction(self):
        """Test with float values between -1 and 0."""
        testCases = [
            (-0.5, 4),  # 8 - (8 * 0.5) = 4
            (-0.25, 6),  # 8 - (8 * 0.25) = 6
            (-0.75, 2)  # 8 - (8 * 0.75) = 2
        ]
        for limitTest, expectedResult in testCases:
            self.runTestWithCPUcount(limitTest, expectedResult)

    def test_negativeInteger(self):
        """Test with integer values <= -1."""
        testCases = [
            (-1, 7),  # 8 - 1 = 7
            (-3, 5),  # 8 - 3 = 5
            (-7, 1)  # 8 - 7 = 1 (minimum value)
        ]
        for limitTest, expectedResult in testCases:
            self.runTestWithCPUcount(limitTest, expectedResult)

    def test_booleanTrue(self):
        """Test with boolean True."""
        self.runTestWithCPUcount(True, 1)
        # Test the specific True case branch
        with patch('multiprocessing.cpu_count', return_value=self.cpuCountMocked):
            resultTest = defineConcurrencyLimit(True)
            self.assertEqual(resultTest, 1)
            # Ensure it's not just coincidentally 1
            self.assertNotEqual(defineConcurrencyLimit(None), 1)

    @patch('Z0Z_tools.parseParameters.oopsieKwargsie')
    def test_stringValues(self, mockOopsieKwargsie):
        """Test string value handling via oopsieKwargsie."""
        testCases = [
            ("true", True, 1),
            ("false", False, self.cpuCountMocked),
            ("none", None, self.cpuCountMocked),
            ("4", 4, 4)
        ]
        for stringInput, oopsieReturn, expectedResult in testCases:
            mockOopsieKwargsie.return_value = oopsieReturn
            self.runTestWithCPUcount(stringInput, expectedResult)

    def test_ensureMinimumOne(self):
        """Test that return value is never less than 1."""
        testCases = [
            (-10, 1),  # Would be -2, but minimum is 1
            (-0.99, 1),  # Would be ~0, but minimum is 1
            (0.1, 1)  # Would be 0.8, but minimum is 1
        ]
        for limitTest, expectedResult in testCases:
            self.runTestWithCPUcount(limitTest, expectedResult)

if __name__ == '__main__':
    unittest.main()


from typing import Optional, Union
import multiprocessing

def defineConcurrencyLimit(limit: Optional[Union[int, float, bool]]) -> int:
    """
    Determine the concurrency limit based on the provided `limit` parameter.
    Parameters:
        limit: The concurrency limit specification. It can be
            - `None`, `False`, or `0`: Use the total number of CPUs.
            - Integer `>= 1`: Directly specifies the concurrency limit.
            - Float `0 < limit < 1`: Fraction of total CPUs to use.
            - Float `-1 < limit < 0`: Subtract a fraction of CPUs from the total.
            - Integer `<= -1`: Subtract the absolute value from total CPUs.
            - `True`: Set concurrency limit to 1.
    Returns:
        concurrencyLimit: The calculated concurrency limit, ensuring it is at least 1.
    """
    cpuTotal = multiprocessing.cpu_count()
    concurrencyLimit = cpuTotal

    if isinstance(limit, str):
        limit = oopsieKwargsie(limit) # type: ignore

    match limit:
        case None | False | 0:
            pass
        case True:
            concurrencyLimit = 1
        case _ if limit >= 1:
            concurrencyLimit = limit
        case _ if 0 < limit < 1:
            concurrencyLimit = int(limit * cpuTotal)
        case _ if -1 < limit < 0:
            concurrencyLimit = cpuTotal - abs(int(limit * cpuTotal))
        case _ if limit <= 1:
            concurrencyLimit = cpuTotal - abs(limit)

    return max(int(concurrencyLimit), 1)

def oopsieKwargsie(huh: str) -> None | str | bool:
    """
    If a calling function passes a `str` to a parameter that shouldn't receive a `str`, `oopsieKwargsie()` can might help you avoid an Exception. It tries to interpret the string as `True`, `False`, or `None`.

    Parameters:
        huh: The input string to be parsed.

    Returns:
        (bool | None | str): The reserved keywords `True`, `False`, or `None` or the original string, `huh`.
    """
    formatted = huh.strip().title()
    if formatted == str(True):
        return True
    elif formatted == str(False):
        return False
    elif formatted == str(None):
        return None
    else:
        return huh

if __name__ == '__main__':
    multiprocessing.set_start_method('spawn')



from pathlib import Path
from unittest.mock import patch, MagicMock
from Z0Z_tools import pipAnything
import tempfile
import unittest
import sys
import subprocess

class TestPipAnything(unittest.TestCase):

    def test_makeListRequirementsFromRequirementsFile(self):
        """
        Test the makeListRequirementsFromRequirementsFile function.
        """
        with tempfile.TemporaryDirectory() as tempdir:
            # Create a temporary requirements file
            requirements_file = Path(tempdir) / 'requirements.txt'
            requirements_content = """
            # This is a comment
            package-A==1.2.3
            package-B>=4.5.6,<=7.8.9
            package_C
            # Another comment
            analyzeAudio@git+https://github.com/hunterhogan/analyzeAudio.git
            """
            requirements_file.write_text(requirements_content)

            # Test with a single file
            requirements = pipAnything.makeListRequirementsFromRequirementsFile(requirements_file)
            self.assertEqual(len(requirements), 4)
            self.assertIn('package-A==1.2.3', requirements)
            self.assertIn('package-B>=4.5.6,<=7.8.9', requirements)
            self.assertIn('package_C', requirements)
            self.assertIn('analyzeAudio@git+https://github.com/hunterhogan/analyzeAudio.git', requirements)

            # Test with multiple files
            requirements2 = pipAnything.makeListRequirementsFromRequirementsFile(requirements_file, requirements_file)
            self.assertEqual(len(requirements2), 4)  # Should still be 4, duplicates removed

            # Test with non-existent file
            nonexistent_file = Path(tempdir) / 'nonexistent.txt'
            requirements3 = pipAnything.makeListRequirementsFromRequirementsFile(nonexistent_file)
            self.assertEqual(len(requirements3), 0)  # Should be empty

    def test_make_setupDOTpy(self):
        """
        Test the make_setupDOTpy function.
        """
        relative_path_package = 'my_package'
        list_requirements = ['numpy', 'pandas']
        setup_content = pipAnything.make_setupDOTpy(relative_path_package, list_requirements)

        # Check if the generated content contains expected elements
        self.assertIn(f"name='{Path(relative_path_package).name}'", setup_content)
        self.assertIn(f"packages=find_packages(where=r'{relative_path_package}')", setup_content)
        self.assertIn(f"package_dir={{'': r'{relative_path_package}'}}", setup_content)
        self.assertIn(f"install_requires={list_requirements},", setup_content)
        self.assertIn("include_package_data=True", setup_content)

    @patch('subprocess.Popen')
    @patch('tempfile.mkdtemp')
    def test_installPackageTarget(self, mockMkdtemp, mockPopen):
        """Test the installPackageTarget function."""
        # Setup mocks
        mockTempDir = str(Path('unittests/dataSamples/tmp/fake_temp_dir').resolve())
        mockMkdtemp.return_value = mockTempDir
        
        mockProcess = MagicMock()
        mockProcess.stdout = ['Installing...', 'Done!']
        mockProcess.wait.return_value = 0
        mockPopen.return_value = mockProcess

        with tempfile.TemporaryDirectory() as tempdir:
            # Create a fake package directory
            package_dir = Path(tempdir) / 'test_package'
            package_dir.mkdir(parents=True, exist_ok=True)
            
            # Create requirements.txt
            requirements_file = package_dir / 'requirements.txt'
            requirements_file.write_text('numpy\npandas')

            # Test installation
            pipAnything.installPackageTarget(package_dir)

            # Verify pip command
            mockPopen.assert_called_once()
            args = mockPopen.call_args[1]['args']
            self.assertEqual(args[0], sys.executable)
            self.assertEqual(args[1:4], ['-m', 'pip', 'install'])

    def test_subtle_jab(self):
        """Test if the code maintains appropriate levels of passive-aggressiveness."""
        with patch('Z0Z_tools.pipAnything.everyone_knows_what___main___is') as mock_main:
            pipAnything.readability_counts()
            mock_main.assert_called_once()
            # If you need this comment to understand what readability_counts() does,
            # then readability_counts() has failed its mission in life

    def test_CLI_functions(self):
        """Test the CLI-related functions."""
        test_cases = [
            # Missing argument
            {
                'argv': ['script.py'],
                'expected_exit': True
            },
            # Invalid directory
            {
                'argv': ['script.py', '/nonexistent/path'],
                'expected_exit': True
            }
        ]

        for case in test_cases:
            with patch('sys.argv', case['argv']), \
                 patch('sys.exit') as mock_exit, \
                 patch('builtins.print') as mock_print:
                
                pipAnything.everyone_knows_what___main___is()
                
                if case['expected_exit']:
                    mock_exit.assert_called_once_with(1)
                    mock_print.assert_called()

    def test_snark_level(self):
        """Ensures the code's snark remains at professionally acceptable levels."""
        originalArgv = sys.argv
        sys.argv = ['script.py']  # Trigger the usage message
        
        with patch('sys.exit') as mock_exit, \
             patch('builtins.print') as mock_print:
            pipAnything.main()
            
            # Verify that the snark was delivered with surgical precision
            mock_print.assert_called()
            mock_exit.assert_called_once_with(1)
            
            # Test if the -m explanation maintains perfect levels of obviousness
            printed_messages = ' '.join(str(call.args[0]) for call in mock_print.call_args_list)
            self.assertNotIn('obviously', printed_messages.lower())
        
        sys.argv = originalArgv  # Restore original argv

    def test_invalid_requirements(self):
        """Test handling of invalid requirements file content."""
        with tempfile.TemporaryDirectory() as tempdir:
            requirements_file = Path(tempdir) / 'requirements.txt'
            invalid_content = """
            invalid==requirement==1.0
            # Comment
            valid-package==1.0
            spaces in package==1.0
            @#$%^invalid characters
            """
            requirements_file.write_text(invalid_content)

            requirements = pipAnything.makeListRequirementsFromRequirementsFile(requirements_file)
            self.assertEqual(len(requirements), 1)
            self.assertIn('valid-package==1.0', requirements)

    def test_multiple_requirements_files(self):
        """Test processing multiple requirements files."""
        with tempfile.TemporaryDirectory() as tempdir:
            # Create two requirements files
            req1 = Path(tempdir) / 'req1.txt'
            req2 = Path(tempdir) / 'req2.txt'
            
            req1.write_text('package-A==1.0\npackage-B==2.0')
            req2.write_text('package-B==2.0\npackage-C==3.0')

            # Test combining requirements
            requirements = pipAnything.makeListRequirementsFromRequirementsFile(req1, req2)
            self.assertEqual(len(requirements), 3)
            self.assertEqual(sorted(requirements), 
                           ['package-A==1.0', 'package-B==2.0', 'package-C==3.0'])

if __name__ == '__main__':
    unittest.main()
